################################################################################
###  This sample uses the alphabet, states, transitions, and tape variables  ###
###  to fully define a Turing machine.  It then uses the tam DSL to create   ###
###  a tile assembly system which self-assembles a representation of that    ###
###  Turing machine running on the input tape.                               ###
###                                                                          ###
###  Note that the Turing machine is assumed to begin simulation with its    ###
###  tape head on the leftmost tape cell of a one-way-infinite-to-the-right  ###
###  tape, and to never move its head past that leftmost cell.               ###
################################################################################

from __future__ import print_function
import tam

N = tam.Direction.North
S = tam.Direction.South
E = tam.Direction.East
W = tam.Direction.West


if __name__ == '__main__':

    #######################################################################################################
    #  In order to simulate a specified Turing machine, the variables 'alphabet', 'states',
    #  'startState', 'acceptState', 'rejectState', 'transitions', and 'tape' should be set
    #  according to the definition of that machine.
    #  The example below creates the tile system which simulates a Turing machine which halts
    #  and accepts for all input strings whose rightmost characters are '10' or '111'.  It
    #  rejects on all other inputs.

    # The tape alphabet of the Turing machine, which must include '_' as the blank symbol
    alphabet = ['0','1','2','3','_']

    # The list of state names of the TM
    states = ['qS', 'q0', 'q1', 'q2', 'q3', 'qA', 'qR']

    startState  = 'qS'  # The state in which the machine starts
    acceptState = 'qA'  # The state into which the machine moves in order to halt and accept
    rejectState = 'qR'  # The state into which the machine moves in order to halt and reject

    # A full transition table mapping every input in every state to an output symbol, direction, and state
    transitions =  {('qS', '1'):('3','R','q0'),
                    ('qS', '0'):('2','R','q0'),
                    ('q0', '0'):('0','R','q0'),
                    ('q0', '1'):('1','R','q0'),
                    ('q0', '_'):('_','L','q1'),
                    ('q1', '0'):('_','R','qA'),
                    ('q1', '1'):('1','L','q2'),
                    ('q1', '2'):('_','R','qA'),
                    ('q1', '3'):('_','R','qR'),
                    ('q2', '0'):('_','R','qR'),
                    ('q2', '1'):('1','L','q3'),
                    ('q2', '2'):('_','R','qR'),
                    ('q2', '3'):('_','R','qR'),
                    ('q3', '0'):('_','R','qR'),
                    ('q3', '1'):('_','R','qA'),
                    ('q3', '2'):('_','R','qR'),
                    ('q3', '3'):('_','R','qA')}

    # A list of the initial tape contents from left to right
    tape = ['1','1','0','0','1','1']
    
    #######################################################################################################
    ##  The following code uses the tam DSL to generate the tile system to simulate the Turing machine
    ##  specified above

    nonHaltingStates = []
    for s in states:
        if s != acceptState and s != rejectState:
            nonHaltingStates.append(s)

    def TuringMachineTransition(state, symbol):
        if transitions.has_key((state,symbol)):
            return transitions[(state,symbol)]
        else:   #  This case just returns a generic set for a "don't care" configuration
            return ('_','R','qS')


    tst = tam.TileSetTemplate()

    left = tam.TileTemplate(name='left')
    right = tam.TileTemplate(name='right')
    rightend = tam.TileTemplate(name='rightend')
    rightnew = tam.TileTemplate(name='rightnew')
    moveleft = tam.TileTemplate(name='moveleft')
    moveright = tam.TileTemplate(name='moveright')
    headfromleft = tam.TileTemplate(name='headfromleft')
    headfromright = tam.TileTemplate(name='headfromright')
    accept = tam.TileTemplate(name='ACCEPT')
    reject = tam.TileTemplate(name='REJECT')
    
    tst.join(1, N, left, left, symbol=alphabet)
    tst.join(1, W, left, left, left=1)
    tst.join(1, W, moveright, left, left=1)
    
    tst.join(1, N, right, right, symbol=alphabet)
    tst.join(1, E, right, right, right=1)
    tst.join(1, E, moveleft, right, right=1)
    tst.join(1, E, right, rightend, right=1)
    
    tst.join(2, E, rightend, rightnew, rightnew=1)
    tst.join(1, N, rightnew, rightend, neednew=1)
    tst.join(1, N, rightend, right, symbol='_')

    tst.join(1, N, moveleft, right, symbol=alphabet)
    tst.join(1, N, moveright, left, symbol=alphabet)
    
    tst.join(1, N, left, headfromright, symbol=alphabet)
    tst.join(1, N, right, headfromleft, symbol=alphabet)

    tst.join(1, E, headfromleft, right, right=1)
    tst.join(1, W, headfromright, left, left=1)
    
    tst.join(1, E, moveright, headfromleft, state=nonHaltingStates)
    tst.join(1, W, moveleft, headfromright, state=nonHaltingStates)
    
    tst.join(2, N, headfromright, moveright, symbol=alphabet, state=nonHaltingStates, move='R')
    tst.join(2, N, headfromright, moveleft, symbol=alphabet, state=nonHaltingStates, move='L')

    tst.join(2, N, headfromleft, moveright, symbol=alphabet, state=nonHaltingStates, move='R')
    tst.join(2, N, headfromleft, moveleft, symbol=alphabet, state=nonHaltingStates, move='L')
    
    tst.join(2, N, headfromright, accept, symbol='_', state='qA', move='R')
    tst.join(2, N, headfromleft, accept, symbol='_', state='qA', move='R')
    
    tst.join(2, N, headfromright, reject, symbol='_', state='qR', move='R')
    tst.join(2, N, headfromleft, reject, symbol='_', state='qR', move='R')
    
    left.addTransition(inputs=('symbol'), outputs=('symbol'), expression='symbol')
    right.addTransition(inputs=('symbol'), outputs=('symbol'), expression='symbol')
    moveright.addTransition(inputs=('state','symbol'), outputs=('state','symbol'), expression='(state,symbol)')
    moveleft.addTransition(inputs=('state','symbol'), outputs=('state','symbol'), expression='(state,symbol)')
    headfromright.addTransition(inputs=('state','symbol'), outputs=('symbol','move','state'), function=TuringMachineTransition)
    headfromleft.addTransition(inputs=('state','symbol'), outputs=('symbol','move','state'), function=TuringMachineTransition)

    left.setLabelFunction(outputs=[], inputs=['symbol'], expression='str(symbol)')
    right.setLabelFunction(outputs=[], inputs=['symbol'], expression='str(symbol)')
    rightend.setLabelFunction(outputs=[], inputs=[], expression='str("_")')
    rightnew.setLabelFunction(outputs=[], inputs=[], expression='str("_")')
    moveleft.setLabelFunction(outputs=[], inputs=['symbol'], expression='str(symbol)')
    moveright.setLabelFunction(outputs=[], inputs=['symbol'], expression='str(symbol)')
    headfromleft.setLabelFunction(outputs=[], inputs=['state','symbol'], expression='"{0},{1}".format(state,symbol)')
    headfromright.setLabelFunction(outputs=[], inputs=['state','symbol'], expression='"{0},{1}".format(state,symbol)')
    accept.setLabelFunction(outputs=[], inputs=[], expression='"ACCEPT"')
    reject.setLabelFunction(outputs=[], inputs=[], expression='"REJECT"')

    tiles = tst.createTiles()

    tileSystem = tam.TileSystem('TM', tiles)
    
    ##  The final section of code is used to generate the initial seed row from which the simulation will begin
    multiSignalType = tam.MultisignalType(symbol=alphabet, state=states, right=1, neednew=1)
    
    if len(tape) == 0:
        tape.append('_')
    tileSystem.addToSeedAssembly((0,0),headfromleft.createTileFromFullInputMultisignal(multiSignalType.createNoNameCheck(symbol=tape[0],state='qS')))
    for i in range(1,len(tape)):
        tileSystem.addToSeedAssembly((i,0),right.createTileFromFullInputMultisignal(multiSignalType.createNoNameCheck(symbol=tape[i],right=1)))
    tileSystem.addToSeedAssembly((len(tape),0),rightend.createTileFromFullInputMultisignal(multiSignalType.createNoNameCheck(neednew=1,right=1)))

    tileSystem.writeToFiles('TM')
