from __future__ import print_function, division

import copy,collections

VERBOSE_SIGNALS = True

"""Contains classes and functions for creating and
manipulating 2D tiles and tile assembly systems in the Tile 
Assembly Model. The fundamental classes are

- Multisignal: a mapping type, mapping signal names to values

- MultisignalType: a mapping type, mapping signal names to lists of valid 
                   values
        
- Tile: a type of tile, composed of one glue on each of the four sides, as well
        as a name, label, and possible other properties such as tilecolor, 
        textcolor, concentration, etc.
            
- TileSystem: a list of unique tiles together with a seed assembly, which is
              a dict mapping 2D coordinates ((int,int) pairs) to Tiles
              
- TileTemplate: a class for specifying a family of Tiles that share a common 
                set of input sides, each with the same GlueTemplate, and a 
                common set of output sides, each with the same GlueTemplate.
                
- TileSetTemplate: a group of TileTemplates and Tiles. This object is
                   responsible for managing the input-output relationships 
                   between the TileTemplates, and can be used to generate a
                   list of all the tiles that are represented by all the
                   TileTemplates (and Tiles) it contains.
"""


##############################################################################
## Exceptions
##############################################################################

class TAMError(Exception):
    """Base class for exceptions produced by the TAM library."""
    pass

class ErrorListError(TAMError):
    """Raised to indicate that one or more errors occurred when executing
    TileSetTemplate.createTiles. It can be queried to retrive that list."""
    def __init__(self, errorList):
        self.errorList = errorList
        
    def errors(self):
        return self.errorList

# Immediate errors are those that ought to result in an excepton immediately
# being raised, since there is no valid reason to want the error to persist
# even temporarily.

# Lazy errors are those that we allow the user to make temporarily, which will
# prevent the tiles from being generated if TileSetTemplate.createTiles is
# called, but which otherwise will be allowed to persist as they construct the
# TileSetTemplate (for instance, in a GUI), so that the errors may be visually
# reported to the user without stopping them from continuing to work.

# immediate
class NonexistentTransitionError(TAMError):
    """Raised to indicate that a transition attempted to be removed does not exist."""
    def __init__(self, outputNamesGiven, transitions):
        self.outputNamesGiven = outputNamesGiven
        self.transitions = transitions
        
    def __str__(self):
        return '{0} is not a set of outputs defining any transition in the list {1}'.format(self.outputNamesGiven, self.transitions)

# immediate when calling create or restrict on a MultisignalType
# lazy when input signal names or output signal names given with transition are invalid
class SignalInvalidNameError(TAMError):
    """Raised to indicate an invalid signal name."""
    def __init__(self, invalidName, validNames):
        self.invalidName = invalidName
        self.validNames = validNames
        
    def __str__(self):
        return '"{0}" is an invalid signal name; valid signal names are {1}'.format(
               self.invalidName, self.validNames)

# immediate when joining or rejoining and MultisignalType objects with overlapping signal names are used
class SignalDuplicateNameError(TAMError):
    """Raised to indicate a duplicate signal name."""
    def __init__(self, name):
        self.name = name
        
    def __str__(self):
        return '"{0}" appears more than once as a signal name'.format(self.name)

# immediate when restricting multisignaltype; 
# lazy when transition outputs invalid value
class SignalInvalidValueError(TAMError):
    """Raised to indicate an invalid signal value."""
    def __init__(self, signalName, invalidVal, validVals):
        self.signalName = signalName
        self.invalidVal = invalidVal
        self.validVals = validVals
       
# immediate when creating or restricting multisignal type
class SignalDuplicateValueError(TAMError):
    """Raised to indicate a duplicate signal name."""
    def __init__(self, vals):
        self.vals = vals
        
    def __str__(self):
        return '"{0}" contains duplicate signal valid values'.format(self.vals)

# immediate; should result in prompting to see if user wants to update other joins
class SignalMissingNameError(TAMError):
    """Raised to indicate a join is missing a signal name for its neighborhood."""
    def __init__(self, neighborhood, join, signalType):
        self.signalType = signalType
        self.neighborhood = neighborhood
        self.join = join
        
    def __str__(self):
        return 'join {0} is missing signal type {1} required for neighborhood {2}'.format(self.join, self.signalType, self.join)

# immediate
class InputOutputSideConflictError(TAMError):
    """Raised to indicate that a single tile template side has been specified as both an input and an output."""
    def __init__(self, tileTemplate, direction):
        self.tileTemplate = tileTemplate
        self.direction = direction
        
    def __str__(self):
        return 'tile template {0} cannot have side {1} as both an input and output'.format(self.tileTemplate, self.direction)

# lazy
class StrengthError(TAMError):
    """Raised to indicate that a join has the incorrect strength, as calculated by the number of input sides on the 
    TileTemplate on the output direction of the join."""
    def __init__(self, tileTemplate, join):
        self.tileTemplate = tileTemplate
        self.join = join
        
    def __str__(self):
        properStrength = 1 if self.tileTemplate.numInputSides() == 2 else 2
        return 'tile template {0} has {1} input sides and must therefore have strength {2} on join {3}'.format(self.tileTemplate, self.tileTemplate.numInputSides(), properStrength, self.join)

# lazy
class TooManyInputSidesError(TAMError):
    """Raised to indicate that a tile template has too many input sides."""
    def __init__(self, tileTemplate):
        self.tileTemplate = tileTemplate
        
    def __str__(self):
        return 'tile template {0} has {1} input sides but must have at most 2'.format(self.tileTemplate, self.tileTemplate.numInputSides())

# lazy
class OutputNotComputedError(TAMError):
    """Raised to indicate that one of the output signals of a TileTemplate was not computed by any transition."""
    
# lazy 
class OutputMultiplyComputedError(TAMError):
    """Raised to indicate that one of the output signals of a TileTemplate has more than one transition computing it."""
    
# lazy because it is only dynamically detectable by calling function with all possible inputs (not all of
# which may have been specified yet if there are remaining TileTemplates left to wire as inputs to this one)
class OutputArityError(TAMError):
    """Raised to indicate that an output tuple from a transition does not have the same length as the list
    of output signal names."""

# immediate because immediately detectable from signature of function
class InputArityError(TAMError):
    """Raised to indicate that the arity of a transition does not have the same length as the list
    of input signal names. If the function takes varargs, then it ensures that the number of input signal
    names specified is at least the number of non-varargs given."""
    
# lazy
class InvalidChooserTileTemplateError(TAMError):
    """Raised to indicate that a tile template or list of tile templates is not an element/subset of the TileTemplates
    that could actually be the output."""
    
# lazy
class MissingChooserError(TAMError):
    """Raised to indicate that the user needs to specify which of multiple TileTemplates to choose for a
    given input multisignal, when more than one could be the output as computed by the library."""

# lazy
class ConflictingChooserError(TAMError):
    """Raised to indicate that there are multiple chooser functions defined for a chooser set on
    a given input multisignal."""

# immediate
class NameDifferenceError(TAMError):
    """Raised to indicate that two multisignals cannot be valueUnion'ed because their signal names are not identical."""
    def __init__(self, names1, names2):
        self.names1 = names1
        self.names2 = names2
        
    def __str__(self):
        return 'list of names {0} must be identical to {1}'.format(self.names1, self.names2)
        
# immediate
class NameOverlapError(TAMError):
    """Raised to indicate that two multisignals cannot be nameUnion'ed because their signal names are not disjoint."""
    def __init__(self, names1, names2):
        self.names1 = names1
        self.names2 = names2
        
    def __str__(self):
        return 'list of names {0} must be disjoint from to {1}'.format(self.names1, self.names2)

class NonexistentChooserSetError(TAMError):
    """Raised to indicate that a chooserSet consisting of the set of tiles does not exist."""
    def __init__(self, tileSet):
        self.tileSet = tileSet
        
    def __str__(self):
        nameList = list([tile.name for tile in self.tileSet])
        return 'There is no chooser set consisting of this list of tile templates: {0}'.format(nameList)

class InputSideMismatchError(TAMError):
    """Raised to indicate that a tileTemplate does not have input sides matching the chooserSet it's being added to"""
    def __init__(self, tileTemplate, tileTemplateInDirs, chooserSetInDirs):
        self.tileTemplate = tileTemplate
        self.tileTemplateInDirs = tileTemplateInDirs
        self.chooserSetInDirs = chooserSetInDirs
        
    def __str__(self):
        return 'TileTemplate {0} has input side list {1} which does not match the list for the chooserSet {2}'.format(self.tileTemplate.name, self.tileTemplateInDirs, self.chooserSetInDirs)

##############################################################################
## These aren't really tile related, but they are useful.
##############################################################################

# allow multi-arity identity; e.g. identity(3,4,5) = (3,4,5)
def identity(*x):
    """Return x, returning a tuple for 0-arity or multi-arity calls.
    
    e.g. identity(3,'a',5) = (3, 'a', 5)"""
    if len(x) == 1:
        return x[0]
    else:
        return x

def reprByProps(obj):
    """Create a string representing obj using the properties. 
    
    Assumes each property can be specified in the constructor.
    
    For instance, an object of type moduleName.A with properties self.a = 123 
    and self.b = 'abc' would have the string 
    "moduleName.A(a=123,b='abc')" returned.
    
    Except for the top-level call, which always performs this pairing, it uses
    the condition of whether obj defines __repr__ to decide whether to use
    repr(val) to represent an attribute value, or to recurse into the property
    and use reprByProps to represent that value also. The idea is that we
    want to use the object's way of representing itself literally if it has
    defined one, but since we use this function to make defining __repr__ 
    easier (and not to replace it), we don't want to perform this check on obj
    itself, which will have __repr__ defined but will rely on reprByProps
    to define it.
    
    WARNING: this should only be used to help define __repr__ in the following
    way:
    
    def __repr__(self):
        return tam.reprByProps(self)
        
    Other than that one use case, it should not be called or used directly.
    
    This is not a serialization mechanism; it will not handle loops in the 
    object graph. It is just a way to make creating "standard" __repr__ 
    methods easier.
    """
    className = obj.__class__.__name__
    nameValList = ', '.join(
        "{0}={1}".format(name, repr(val) if hasattr(val, '__repr__') else reprByProps(val))
        for name, val in sorted(obj.__dict__.items())
    )
#    this code was an attempt to use only those properties that the constructor
#      specifies through arguments, but it has some issues that I forget
#    nameValList = ', '.join(
#        "{0}={1}".format(name, repr(obj.__dict__[name]) if hasattr(obj.__dict__[name],'__repr__') else reprByProps(obj.__dict__[name]))
#        for name in sorted(inspect.getargspec(obj.__init__).args)
#        if name != 'self'
#    )
    moduleName = obj.__class__.__module__
    return '{moduleName}.{className}({nameValList})'.format(
        moduleName=moduleName, className=className, nameValList=nameValList)


# taken from http://code.activestate.com/recipes/413486/
def Enum(*names):
    assert names, "Empty enums are not supported" # <- Don't like empty enums? Uncomment!

    class EnumClass(object):
        __slots__ = tuple(names) + ('__dict__',)
        def __iter__(self):        return iter(constants)
        def __len__(self):         return len(constants)
        def __getitem__(self, i):  return constants[i]
        def __repr__(self):        return 'Enum' + str(names)
        def __str__(self):         return 'enum ' + str(constants)

    class EnumValue(object):
        __slots__ = ('__value')
        def __init__(self, value): self.__value = value
        Value = property(lambda self: self.__value)
        EnumType = property(lambda self: EnumType)
        def __hash__(self):        return hash(self.__value)
        def __cmp__(self, other):
            # C fans might want to remove the following assertion
            # to make all enums comparable by ordinal value {;))
            assert self.EnumType is other.EnumType, "Only values from the same enum are comparable"
            return cmp(self.__value, other.__value)
        def __invert__(self):      return constants[maximum - self.__value]
        def __nonzero__(self):     return bool(self.__value)
        def __repr__(self):        return str(names[self.__value])

    maximum = len(names) - 1
    constants = [None] * len(names)
    for i, each in enumerate(names):
        val = EnumValue(i)
        setattr(EnumClass, each, val)
        constants[i] = val
    constants = tuple(constants)
    EnumType = EnumClass()
    return EnumType

Direction = Enum('East', 'North', 'South', 'West', 'Nondet')
def oppositeDirection(self):
    """Gets opposite direction of self; raises ValueError if self is Nondet."""
    if self == Direction.East: return Direction.West
    if self == Direction.West: return Direction.East
    if self == Direction.North: return Direction.South
    if self == Direction.South: return Direction.North
    if self == Direction.Nondet: raise ValueError('there is no opposite of Direction.Nondet')

def directionShortName(self):
    if self == Direction.East: return 'E'
    if self == Direction.West: return 'W'
    if self == Direction.North: return 'N'
    if self == Direction.South: return 'S'
    if self == Direction.Nondet: return 'Nondet'

N = Direction.North
S = Direction.South
E = Direction.East
W = Direction.West
NONDET = Direction.West

# Usage of Enum:
#
#if __name__ == '__main__':
#    print('\n*** Enum Demo ***')
#    print('--- Days of week ---')
#    Days = Enum('Mo', 'Tu', 'We', 'Th', 'Fr', 'Sa', 'Su')
#    print(Days)
#    print(Days.Mo)
#    print(Days.Fr)
#    print(Days.Mo < Days.Fr)
#    print(list(Days))
#    for each in Days:
#        print('Day:', each)
#    print('--- Yes/No ---')
#    Confirmation = Enum('No', 'Yes')
#    answer = Confirmation.No
#    print('Your answer is not {0}'.format(not answer))


##############################################################################
## Tile stuff below here.
##############################################################################

class Tile:
    """
    Represents a tile type, with the properties that are recognized by
    Matt Patitz's TAS application.
    """
    def __init__(self, name, label='',
                 tilecolor='white', textcolor='black', concentration=1,
                 northglue=('', 0), southglue=('', 0),
                 westglue=('', 0), eastglue=('', 0)):
        """
        Each glue is a tuple (label, strength).
        """
        self.name = name
        self.label = label
        self.tilecolor = tilecolor
        self.textcolor = textcolor
        self.concentration = concentration
        self.northglue = northglue
        self.southglue = southglue
        self.westglue = westglue
        self.eastglue = eastglue
    
    def __eq__(self, other):
        return all(self.__dict__[propname] == other.__dict__[propname] 
                   for propname in self.__dict__.keys())
    
    def __ne__(self, other):
        return not (self == other)

    def tdsFormat(self):
        """
        Returns a string in the format recognized by the TAS application,
        suitable for writing to a .TDS file.
        """
        return """TILENAME %s
LABEL %s
TILECOLOR %s
TEXTCOLOR %s
CONCENTRATION %s
NORTHBIND %d
SOUTHBIND %d
WESTBIND %d
EASTBIND %d
NORTHLABEL %s
SOUTHLABEL %s
WESTLABEL %s
EASTLABEL %s
CREATE""" % (self.name, self.label, self.tilecolor, self.textcolor,
             self.concentration,
             self.northglue[1], self.southglue[1],
             self.westglue[1], self.eastglue[1],
             self.northglue[0], self.southglue[0],
             self.westglue[0], self.eastglue[0])

    def __str__(self):
        return self.tdsFormat()
    
    def __repr__(self):
        return reprByProps(self)
    
    def reflectNS(self, axis):
        """Reflect tile north/south."""
        return Tile(
            name=self.name,
            label=self.label,
            tilecolor=self.tilecolor,
            textcolor=self.textcolor,
            concentration=self.concentration,
            northglue=self.southglue,
            southglue=self.northglue,
            westglue=self.westglue,
            eastglue=self.eastglue
        )
        
    def reflectEW(self, axis):
        """Reflect tile east/west."""
        return Tile(
            name=self.name,
            label=self.label,
            tilecolor=self.tilecolor,
            textcolor=self.textcolor,
            concentration=self.concentration,
            northglue=self.northglue,
            southglue=self.southglue,
            westglue=self.eastglue,
            eastglue=self.westglue
        )
    
    def rotateLeft(self):
        """Rotate tile left."""
        return Tile(
            name=self.name,
            label=self.label,
            tilecolor=self.tilecolor,
            textcolor=self.textcolor,
            concentration=self.concentration,
            northglue=self.eastglue,
            southglue=self.westglue,
            westglue=self.northglue,
            eastglue=self.southglue
        )
        
    def rotateRight(self):
        """Rotate tile left."""
        return Tile(
            name=self.name,
            label=self.label,
            tilecolor=self.tilecolor,
            textcolor=self.textcolor,
            concentration=self.concentration,
            northglue=self.westglue,
            southglue=self.eastglue,
            westglue=self.southglue,
            eastglue=self.northglue
        )
    
    def rotate180(self):
        """Rotate tile 180 degrees.
        
        Note that this is different from reflecting since sides along both axes
        are swapped."""
        return Tile(
            name=self.name,
            label=self.label,
            tilecolor=self.tilecolor,
            textcolor=self.textcolor,
            concentration=self.concentration,
            northglue=self.southglue,
            southglue=self.northglue,
            westglue=self.eastglue,
            eastglue=self.westglue
        )
    
    
class TileSystem:
    """
    Represents a tile assembly system. This is a set of tile types, together with
    all other contextual information (such as the seed assembly and temperature)
    needed to simulate the self assembly of the tiles.
    """
    def __init__(self, name, tileTypes=[], seedAssembly={}):
        """Create a tile assembly system with the given name, no tile types,
        and an empty seed assembly."""
        self.name = name        # name of the tile assembly system
        self.tileTypes = tileTypes    # set of tile types
        self.seedAssembly = seedAssembly  # maps position in Z^3 to tile type
    
    def tdpFormat(self):
        """TDP format of this tile assembly system.
        
        Return a string representing the tile assembly system's .TDP file, suitable
        for creating a file that can be read by the TAS application.
        """
        seedTileNamesAndPositions = (
            "{0} {1}".format(self.seedAssembly[pos].name, ' '.join(map(str, pos)))
            for pos in self.seedAssembly.keys())
        return self.name + ".tds\n" + '\n'.join(seedTileNamesAndPositions)

    def tdsFormat(self):
        """TDS format of tile set of this tile assembly system.
        
        Return a string representing all the tiles in the format recognized 
        by the TAS application, suitable for writing to a .TDS file.
        """
        return '\n\n'.join(map(Tile.tdsFormat, self.tileTypes))

    def __str__(self):
        return self.tdpFormat()

    def __repr__(self):
        return reprByProps(self)

    def addTileType(self, tile):
        """Add tile to this tile assembly system's set of tiles."""
        #self.tileTypes.add(tile)
        #self.tileTypes.append(tile)
        self.addTileTypes([tile])

    def addTileTypes(self, tiles):
        """Add tiles to this tile assembly system's set of tiles."""
        #self.tileTypes |= set(tiles)
        #existingNames = {tile.name for tile in self.tileTypes} #PYTHON3
        existingNames = set([tile.name for tile in self.tileTypes])
        for name in (tile.name for tile in tiles):
            if name in existingNames:
                raise ValueError('tile with name {0} already exists'.format(name))
            existingNames.add(name)
        self.tileTypes += tiles
        
    def addToSeedAssembly(self, pos, tile):
        """Add tile to 3D position represented by pos."""
        self.seedAssembly[pos] = tile
    
    def writeToFiles(self, outFilename):
        """Write out tile assembly system to files for use by TAS program.
        
        Write the given tileSystem to the files named <outFilename>.tdp
        (for tile system) and <outFilename>.tds (for tile set), in the format
        recognized by the TAS application.
        """
        tds = open(outFilename + '.tds', 'w')
        tds.write(self.tdsFormat())
        tds.close()
        tdp = open(outFilename + '.tdp', 'w')
        tdp.write(self.tdpFormat())
        tdp.close()

def isSequence(seq):
#    return type(seq) in (tuple, list)
    return isinstance(seq, tuple) or isinstance(seq, list)

class MultisignalType:
    """Tuple of signal types.
    
    Each signal type is a pair (name, validValues), where name is a string 
    representing the name of the signal and valid values is a list of strings
    representing the values the signal can take."""
    def __init__(self, verbose=VERBOSE_SIGNALS, **signalTypeDict):
        self.signalTypeDict = dict(signalTypeDict)
        for name, values in self.signalTypeDict.items():
            if not isSequence(values):
                self.signalTypeDict[name] = (values,)
            else:
                self.signalTypeDict[name] = tuple(values)
        self.updateSignalTypeList()
        self.verbose = verbose
        for validVals in self.signalTypeDict.values():
            if len(set(validVals)) != len(validVals):
                raise SignalDuplicateValueError(validVals)
       
    def noChoiceMultisignal(self):
        """Returns the multisignal consisting of those signals which have only one valid value"""

        ret = Multisignal([(signalName, signalValues[0]) for (signalName,signalValues) in self.signalTypeDict.items() 
                               if len(signalValues) == 1])
        return ret
       
    def __len__(self):
        return len(self.signalTypeDict)

    def names(self):
        """Return list of signal names."""
        return [name for (name, values) in self.signalTypeList]
    
    @classmethod
    def empty(verbose=VERBOSE_SIGNALS):
        return MultisignalType(**dict())
            
    def clone(self):
        return MultisignalType(self.verbose, **dict(self.signalTypeDict))
        
    def updateSignalTypeList(self):
        self.signalTypeList = list(self.signalTypeDict.items())
        self.signalTypeList.sort(key=lambda (name, vals): name)
        
    def addSignalTypes(self, **newSignalTypeDict):
        """Add new signal types."""
        for validVals in newSignalTypeDict.values():
            if len(set(validVals)) != len(validVals):
                raise SignalDuplicateValueError(validVals)
        for name, validVals in newSignalTypeDict.items():
            if name in self.signalTypeDict:
                raise SignalDuplicateNameError(name)
            self.signalTypeDict[name] = tuple(validVals)
            self.signalTypeList.append((name, validVals))
        self.signalTypeList.sort(key=lambda (name, vals): name)
        
    def removeSignalTypes(self, *names):
        """Remove signal types with the given names."""
        for name in names:
            if name not in self.signalTypeDict:
                raise SignalInvalidNameError(name, self.signalTypeDict.keys())
            del self.signalTypeDict[name]
        self.signalTypeList = list(self.signalTypeDict.items())
        self.signalTypeList.sort(key=lambda (name, vals): name)
        
    def create(self, **nameValDict):
        """Create a multisignal with the specified mapping of names to values."""
        for name, val in nameValDict.items():
            if name not in self.signalTypeDict:
                raise SignalInvalidNameError(name, self.signalTypeDict.keys())
            validVals = self.signalTypeDict[name]
            if val not in validVals:
                raise SignalInvalidValueError(name, val, validVals)
        nameValList = [(name, nameValDict[name]) for (name, validVals) in self.signalTypeList]
        return Multisignal(nameValList, self.verbose)

    def createNoNameCheck(self, **nameValDict):
        """Create a multisignal with the specified mapping of names to values.
        
        The names are not checked, so the given nameValDict could contain
        more or fewer signal names than this MultisignalType represents."""
        for name, val in nameValDict.items():
            if name in self.signalTypeDict:
                validVals = self.signalTypeDict[name]
                if val not in validVals:
                    raise SignalInvalidValueError(name, val, validVals)
        nameValList = [(name, nameValDict[name]) for (name, validVals) in self.signalTypeList
                       if name in nameValDict]
        return Multisignal(nameValList, self.verbose)

    def restrict(self, **nameValuesDict):
        """Create a MultisignalType with the given restrictions of values.
        
        Assume for any unspecified signal name that all original values are kept."""
        for name, values in nameValuesDict.items():
            if not isSequence(values): 
                nameValuesDict[name] = values = (values,)
            if name not in self.signalTypeDict:
                raise SignalInvalidNameError(name, self.signalTypeDict.keys())
            validVals = self.signalTypeDict[name]
            for value in values:
                if value not in validVals:
                    raise SignalInvalidValueError(name, value, validVals)
        for name, values in self.signalTypeDict.items():
            if name not in nameValuesDict:
                nameValuesDict[name] = values
        return MultisignalType(verbose=self.verbose, **nameValuesDict)
        
    def __getitem__(self, name):
        """Allows multisignals to index values by name; i.e. ms['bit'] == ['0','1']"""
        return self.signalTypeDict[name]

    def __setitem__(self, name, validVals):
        """Allows multisignals to alter valid values by name; i.e. ms['bit'] = ['0','1']"""
        self.signalTypeDict[name] = tuple(validVals)
        self.updateSignalTypeList()

    def __delitem__(self, name):
        """Allows multisignals to delete values by name; i.e. del(ms['bit'])"""
        del(self.signalTypeDict[name])
        self.signalTypeList = [(sigName, validVals) 
                               for (sigName, validVals) in self.signalTypeList 
                               if name != sigName]

    def __str__(self):
        return ';'.join('{0}:{1}'.format(name, validVals) for name, validVals in self.signalTypeList)
    
    def __repr__(self):
        return (("tam.MultisignalType(verbose={v}, "
             + "**{signalTypeDict})").format(
               signalTypeDict=self.signalTypeDict, v=self.verbose))

    def __iter__(self):
        return self.multisignals()
    
    def multisignals(self):
        """Enumerate all possible multisignals represented by this type."""
        lengths = [len(values) for name, values in self.signalTypeList]
        curSignalIndices = [0] * len(self.signalTypeList)
        numGlues = 1
        for length in lengths:
            numGlues *= length
        for glueIdx in range(numGlues):
            multisignalDict = dict((signalName, signalsList[i]) for i, (signalName, signalsList)
                    in zip(curSignalIndices, self.signalTypeList))
            yield self.create(**multisignalDict)
            # increment to next glue using carry addition
            for i in range(len(self.signalTypeList)):
                curSignalIndices[i] += 1
                # update next index if necessary, otherwise leave rest alone
                if curSignalIndices[i] == lengths[i]:
                    curSignalIndices[i] = 0
                else:
                    break
                
    def __eq__(self, other):
        if set(self.signalTypeDict.keys()) != set(other.signalTypeDict.keys()):
            return False
        for ((name, values), (otherName, otherValues)) in zip(self.signalTypeList, other.signalTypeList):
            if values != otherValues:
                return False
        return True


    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(tuple(self.signalTypeList))
    
    def nameUnion(self, other):
        """Combine signal types with disjoint names to be the union of the signal types.
        
        For instance,
        
            carryMst = tam.MultisignalType(carry=['n','c'])
            borrowMst = tam.MultisignalType(borrow=['nb','b'])
            carryBorrowMst = carryMst.nameUnion(borrowMst)
        
        creates a MultisignalType that can represent a carry and a borrow
        simultaneously. They must also have disjoint signal names.
        """
        for name, validVals in self.signalTypeDict.items():
            if name in other.signalTypeDict:
                raise NameOverlapError(list(self.signalTypeDict.keys()), list(other.signalTypeDict.keys()))
        signalTypeDict = dict(self.signalTypeDict)
        signalTypeDict.update(other.signalTypeDict)
        newMst = MultisignalType(verbose=self.verbose, **signalTypeDict)
        return newMst
    
    def valueUnion(self, other):
        """Combine signal types with the same names by taking the union of their
        values for each shared name.
        
        For instance,
        
            carryMst = tam.MultisignalType(carry=['n','c'])
            borrowMst = tam.MultisignalType(carry=['n','b'])
            carryBorrowMst = carryMst.valueUnion(borrowMst)
        
        creates a MultisignalType that can represent a carry with values
        ['n','b','c']. They must also have exactly the same signal names.
        """
        if set(self.signalTypeDict.keys()) != set(other.signalTypeDict.keys()):
            raise NameDifferenceError(list(self.signalTypeDict.keys()), list(other.signalTypeDict.keys()))
        signalTypeDict = dict(self.signalTypeDict)
        for name in signalTypeDict:
            signalTypeDict[name] = list(set(signalTypeDict[name] + other.signalTypeDict[name]))
        newMst = MultisignalType(verbose=self.verbose, **signalTypeDict)
        return newMst
    
    def isValidMultisignal(self, ms):
        """Indicates whether ms is a Multisignal that this MultisignalType produces."""
        if set(ms.nameValDict.keys()) != set(self.signalTypeDict.keys()):
            return False
        for name, value in ms.nameValDict.items():
            if value not in self.signalTypeDict[name]:
                return False
        return True
    

class Multisignal:
    """Tuple of signals.
    
    Do not use the constructor; use MultisignalType.createMultisignal"""
    def __init__(self, nameValList, verbose=VERBOSE_SIGNALS):
        self.nameValDict = dict(nameValList)
        self.nameValList = sorted(nameValList)
        self.verbose = verbose
        
    def __str__(self):
        """Return string in a format suitable for using as a glue label."""
        if self.verbose:
            return ','.join('{0}={1}'.format(name, val) for name, val in self.nameValList)
        else:
            return ','.join('{0}'.format(val) for name, val in self.nameValList)
    
    def __len__(self):
        return len(self.nameValDict)

    def __repr__(self):
        return (("tam.Multisignal(nameValList={nameValList}, "
             + "verbose={verbose})").format(nameValList=self.nameValList,
                                           verbose=self.verbose))
    
    def __getitem__(self, name):
        """Allows multisignals to index values by name; i.e. ms['bit'] == '1'"""
        return self.nameValDict[name]

    def __setitem__(self, name, val):
        """Allows multisignals to alter values by name; i.e. ms['bit'] = '0'"""
        oldval = self.nameValDict[name]
        self.nameValDict[name] = val
        self.nameValList[self.nameValList.index((name, oldval))] = val

    def __delitem__(self, name):
        """Allows multisignals to delete values by name; i.e. del(ms['bit'])"""
        del(self.nameValDict[name])
        self.nameValList = [(sigName, val) 
                            for (sigName, val) in self.nameValList 
                            if name != sigName]

    def __eq__(self, other):
        """Indicate if these represent the same name/value pairs; ignores verbosity."""
        return self.nameValList == other.nameValList
        
    def __ne__(self, other):
        return not (self == other)
    
    def __hash__(self):
        return hash(tuple(self.nameValList))
    
    def __add__(self, other):
        selfNameSet = set(name for (name,value) in self.nameValList)
        otherNameSet = set(name for (name,value) in other.nameValList)
        if not selfNameSet.isdisjoint(otherNameSet):
            raise SignalDuplicateNameError(selfNameSet & otherNameSet)
        nameValList = self.nameValList + other.nameValList
        nameValList.sort(key=lambda (name, value): name)
        return Multisignal(nameValList, verbose=self.verbose)
    
    def nameUnionDuplicateSignalAllowed(self, other):
        selfNameSet = set(name for (name,value) in self.nameValList)
        otherNameSet = set(name for (name,value) in other.nameValList)
        if not selfNameSet.isdisjoint(otherNameSet):
            for name in (selfNameSet & otherNameSet):
                if self.nameValDict[name] != other.nameValDict[name]:
                    raise ValueError('{0} must be exactly equal to {1} to combine these Multisignals'.format(self.nameValDict[name],other.nameValDict[name]))
            nameValDict = dict(self.nameValDict)
            nameValDict.update(other.nameValDict)
            nameValList = list(nameValDict.items())
        else:
            nameValList = self.nameValList + other.nameValList
        nameValList.sort(key=lambda (name, value): name)
        return Multisignal(nameValList, verbose=self.verbose)
    
    def __iter__(self):
        """Iterate over (name,value) pairs of this multisignal."""
        return iter(self.nameValList)
    






class GlueTemplate:
    """A GlueTemplate represents a "type of glue".
    
    A GlueTemplate specifies a binding strength and a MultisignalType.
    
    It is for binding tiles in specific positions in an assembly, and 
    encompasses a collection of related glues.
    It is not intended for direct use by client code, but is used to configure
    signal passing between TileTemplates by the tam library.
    
    A GlueTemplate is added as either an input side or an output side to a
    TileTemplate; while nothing prevents doing otherwise, it only makes sense
    to add a GlueTemplate as an input side to at least one TileTemplate and as
    an output side to at least one TileTemplate (not necessarily a different
    TileTemplate).
    """
    def __init__(self, strength=1, multisignalType=MultisignalType(verbose=False)):
        self.strength = strength
        self.multisignalType = multisignalType
    
    def create(self, **nameValDict):
        """Converts a name/value dict to a glue, a pair (label,strength).

        The multisignal is presented as a keyword argument list.

        The form is 'name1=value1,name2=value2,...' if property names are
        enabled for this TileTemplate; otherwise the format is
        value1,value2,..."""
        return (str(self.multisignalType.create(**nameValDict)), self.strength)
    
    def createLabel(self, **nameValDict):
        """Converts a dict of name:value pairs to a string representing the 
        glue label.

        The form is 'name1=value1,name2=value2,...' if property names are
        enabled for this TileTemplate; otherwise the format is
        value1_value2..."""
        return (self.create(**nameValDict))[0]
    
    def __str__(self):
        return '[strength={0}, multisignalType={1}]'.format(
            self.strength, self.multisignalType)

    def __repr__(self):
        return reprByProps(self)

    def __iter__(self):
        """Enumerate all possible glues, represented as (string,int) pairs,
        for this template. In the __add__ example above,

            for glue in embedCarryAndBorrowGlue:
                print(glue)

        prints the four glues

            (n_nb,1)
            (n_b,1)
            (c_nb,1)
            (c_b,1)
        """
        return self.glues()

    def glues(self):
        """
        Used by __iter__ to create the generator for enumerating all possible
        glue labels (as a string), given the list of lists of signals.
        """
        for multisignal in self.multisignalType:
            yield (str(multisignal), self.strength)

class Transition:
    """Represents a signal transition for computing output signals
    of a tile from input signals."""
    def __init__(self, inputs=(), outputs=(), function=None, expression=None, table=None):
        """Add a transition computing output signal(s) to this TileTemplate.
        
        inputs and outputs are lists of strings specifying the names of the input
        and output signal names, in the order in which the function transition 
        receives them as arguments or returns them in a tuple."""
        self.function = ensureIsFunction(inputs, function, expression, table)
        self.inputNames = inputs
        self.outputNames = outputs
    
    def __call__(self, *args):
        """Call this transition's function."""
        return self.function(*args)
    
    def apply(self, inputMultisignal):
        """Call this transition's function with the given input multisignal
        and return an output multisignal.
        
        The difference between this and __call__ is the preservation of input
        and output names instead of positional parameters for input and 
        output tuples for output."""
        inputValues = [inputMultisignal[inputName] for inputName in self.inputNames]
        # compute output from input
        outputValues = self.function(*inputValues)
        # if outputs is None, propagate that
        if outputValues is None:
            return None
        # place output in a tuple if it is not a sequence already
        if not isSequence(outputValues):
            outputValues = (outputValues,)
        outputNameValueList = [(self.outputNames[pos], outputValue) 
                               for (pos,outputValue) in enumerate(outputValues)]
        outputMultisignal = Multisignal(outputNameValueList, verbose=inputMultisignal.verbose)
        return outputMultisignal


class PropertyFunction:
    """Represents a function for computing non-glue properties of a tile from
    either output and/or input signals."""
    def __init__(self, inputs=(), outputs=(), function=None, expression=None):
        """Add a transition computing output signal(s) to this TileTemplate.
        
        inputs and outputs are lists of strings specifying the names of the input
        and output signal names, in the order in which the function transition 
        receives them as arguments or returns them in a tuple.
        
        Note that names in expression can refer only to signal names that are
        contained in exactly one of inputNames or outputNames. Otherwise, the
        positional ordering of inputs to function must be used, in which
        function will be called with the values associated with inputNames
        given first, followed by those in outputNames."""
        if not isSequence(inputs):
            inputs = (inputs,)
        if not isSequence(outputs):
            outputs = (outputs,)
        inputNamesSet = set(inputs)
        outputNamesSet = set(outputs)
        if not (inputNamesSet.isdisjoint(outputNamesSet)):
            raise SignalDuplicateNameError(inputNamesSet & outputNamesSet)
        
        self.function = ensureIsFunction(inputs + outputs, function, expression, None)        
        self.inputNames = inputs
        self.outputNames = outputs
    
    def apply(self, inputMultisignal, outputMultisignal):
        """Call this transition's function with the given input and output
        multisignals and return the value of the property."""
        inputValues = [inputMultisignal[inputName] for inputName in self.inputNames]
        outputValues = [outputMultisignal[outputName] for outputName in self.outputNames]
        # compute output from input
        propertyValue = self.function(*(inputValues + outputValues))
        return propertyValue

class ChooserFunction:
    """Represents a function for computing which tileset should be generated from
    within a chooserSet for a given input multisignal."""
    def __init__(self, inputNames=(), function=None, expression=None, table=None):
        """Add a function for selecting the appropriate tileTemplate from a chooserSet.
        
        inputs and outputs are lists of strings specifying the names of the input
        and output signal names, in the order in which the function transition 
        receives them as arguments or returns them in a tuple.
        
        Note that names in expression can refer only to signal names that are
        contained in exactly one of inputNames or outputNames. Otherwise, the
        positional ordering of inputs to function must be used, in which
        function will be called with the values associated with inputNames
        given first, followed by those in outputNames."""
        if not isSequence(inputNames):
            inputNames = (inputNames,)

        self.function = ensureIsFunction(inputNames, function, expression, table)        
        self.inputNames = inputNames
    
    def apply(self, inputMultisignal):
        """Call this transition's function with the given input and output
        multisignals and return the value of the property."""
        inputValues = [inputMultisignal[inputName] for inputName in self.inputNames]
        # compute output from input
        tileNames = ensureIsSequence(self.function(*(inputValues)))
        return tileNames
    

def expressionToFunction(expression, inputNames):
    """Turns an expression as a string, with a list of input names, into a 
    function taking parameters with those names as inputs, which returns the
    value of the expression."""
    def function(*vargs):
        z = zip(inputNames, vargs)
        d = dict(z)
        return eval(expression, {}, dict(zip(inputNames, vargs)))
    return function


def tableToFunction(table):
    """Turns a table of input/output pairs of tuples (implemented either as a 
    list of pairs of input/output tuples, or a dict mapping input tuples to 
    output tuples) into a function taking parameters, which returns the
    value of the output tuple, or raises a ValueError if the input tuple is not
    in the table."""
    if isinstance(table, list):
        table = dict(table)
    def function(*vargs):
        return table[tuple(vargs)]
    return function

def ensureIsFunction(inputNames, function, expression, table):
    numNotNone = 0
    if expression is not None:
        retFunction = expressionToFunction(expression, inputNames)
        numNotNone += 1
    if table is not None:
        retFunction = tableToFunction(table)
        numNotNone += 1
    if function is not None:
        retFunction = function
        numNotNone += 1
    if numNotNone != 1:
        raise ValueError('exactly one of function, expression, or table must be specified')
    return retFunction

class TileTemplate:
    """Represents a "kind of tile type", a family of related tile types.

    It can only be used for specifying a collection of related tiles
    that all share the same input/output/terminal side designations.
    
    A TileTemplate consists of 1-4 (GlueTemplates, direction) pairs, one for
    each of the north, south, east, and west sides of the tile type,
    specifying the type of glue that goes on that side and whether it is an
    input or output side (sides without glue are terminal). GlueTemplates are
    not intended to be controlled by the client, but by a TileSetTemplate
    object, which handles the input/output relationships between TileTemplate
    objects. A TileTemplate also requires
    functions specifying how signals from the input sides create signals on
    the output sides. Any direction not added as an input side or output side
    is implicitly a terminal side, and will result in strength 0 glues being
    produced on that side. Each tile type produced by this template will be
    given a unique name and label based on the input side glues.
    """
    
    def __init__(self, name, **tileTypeArgs):
        """Create a TileTemplate with all terminal sides.

        tileTypeArgs is a dict specifying extra keyword arguments to be passed
        to the TileType constructor when creating tile types; for instance,
        tilecolor or textcolor. It should not contain any of the glues or 
        label or name, as these are computed from the input signals. This means
        that TileTemplates can be fed these arguments, e.g.:
          tileTemplate = TileTemplate(name='start',tilecolor='green',textcolor='orange')
        """
        self.name = name # a special prefix to attach to all tile type names
        self.inputJoins = []  # list of Joins
        self.outputJoins = [] # list of Joins
        self.transitions = [] # list of triples (func, inputs, outputs)
        self.propertyFunctions = {} # dict mapping property name to triple (func, inputs, outputs)
        self.tileTypeArgs = tileTypeArgs # dict giving tile type properties to hardcode, e.g.  {'label': 0}
        self.auxiliaryInputMultisignalType = MultisignalType.empty()
        
    def addAuxiliaryInput(self, multisignalType):
        self.auxiliaryInputMultisignalType = self.auxiliaryInputMultisignalType.nameUnion(multisignalType)

    def inputDirMultisignalTypeDict(self):
        """Return dict mapping (direction, MultisignalType), including Nondet, 
        of input multisignal types."""
        ret = collections.defaultdict(MultisignalType)
        for join in self.inputJoins:
            direction = oppositeDirection(join.neighborhood.direction)
            if direction in ret:
                ret[direction] = ret[direction].valueUnion(join.multisignalType)
            else:
                ret[direction] = join.multisignalType
        if self.auxiliaryInputMultisignalType:
            ret[Direction.Nondet] = self.auxiliaryInputMultisignalType
        return ret
    
    def outputDirMultisignalTypeDict(self):
        """Return dict mapping (direction, MultisignalType), 
        of output multisignal types."""
        ret = collections.defaultdict(MultisignalType)
        for join in self.outputJoins:
            direction = join.neighborhood.direction
            if direction in ret:
                ret[direction] = ret[direction].valueUnion(join.multisignalType)
            else:
                ret[direction] = join.multisignalType

        return ret
    
    def inputDirMultisignalTypeList(self):
        """Return list of pairs (direction, MultisignalType), in alphabetical
        order of direction, including Nondet, of input multisignal types."""
        ret = self.inputDirMultisignalTypeDict()
        ret = list(ret.items())
        ret.sort()
        return ret
    
    def outputDirMultisignalTypeList(self):
        """Return list of pairs (direction, MultisignalType), in alphabetical
        order of direction, of output multisignal types."""
        ret = self.outputDirMultisignalTypeDict()
        ret = list(ret.items())
        ret.sort()
        return ret

    def isInputSide(self, direction):
        for join in self.inputJoins:
            if join.neighborhood.direction == oppositeDirection(direction):
                return True
        return False

    def inputNeighborhood(self, direction):
        for join in self.inputJoins:
            if join.neighborhood.direction == oppositeDirection(direction):
                return join.neighborhood
        raise ValueError('no input neighborhood in direction {0}'.format(direction))

    def inputNeighborhoods(self):
        retDict = {}
        for direction in (N,S,W,E):
            for join in self.inputJoins:
                if join.neighborhood.direction == oppositeDirection(direction):
                    retDict[direction] = join.neighborhood
        return retDict

    def isOutputSide(self, direction):
        for join in self.outputJoins:
            if join.neighborhood.direction == direction:
                return True
        return False

    def outputNeighborhood(self, direction):
        for join in self.outputJoins:
            if join.neighborhood.direction == direction:
                return join.neighborhood
        raise ValueError('no output neighborhood in direction {0}'.format(direction))

    def createTilesFromInputMultisignal(self, inputMultisignal):
        if self.auxiliaryInputMultisignalType:
            return [self.createTileFromFullInputMultisignal(inputMultisignal + auxiliaryInputMultisignal) for auxiliaryInputMultisignal in self.auxiliaryInputMultisignalType]
        else:
            return [self.createTileFromFullInputMultisignal(inputMultisignal)]
    
    def createTileFromFullInputMultisignal(self, inputMultisignal):
        """Create a tile from the input multisignal provided."""
        inputDirMultisignalTypeList = self.inputDirMultisignalTypeList()   

        separator = '_' if len(inputMultisignal.nameValDict) > 0 else ''
        tilename = self.name + separator + '_'.join('{0}:{1}'.format(
            direction, multisignalType.createNoNameCheck(**inputMultisignal.nameValDict))
            for (direction, multisignalType) in inputDirMultisignalTypeList
        )
        northglue,southglue,eastglue,westglue = [None]*4
        
        # set input glues
        for direction,multisignalType in inputDirMultisignalTypeList:
            glueLabel = str(multisignalType.createNoNameCheck(**inputMultisignal.nameValDict))
            inputNeighborhood = self.inputNeighborhood(direction)
            glueAnnotation = inputNeighborhood.glueAnnotation
            glue = (glueLabel + glueAnnotation, inputNeighborhood.strength)
            if   direction == N: northglue = glue
            elif direction == S: southglue = glue
            elif direction == W: westglue = glue
            elif direction == E: eastglue = glue
            elif direction == NONDET: pass
            else: raise ValueError('{0} is not a valid direction'.format(direction))

        outputMultisignal = Multisignal({})
        
        # populate outputMultisignal
        for transition in self.transitions:
            outputMultisignal += transition.apply(inputMultisignal)
        
        # set output glues
        completeOutputMultisignal = Multisignal({})
        outputDirMultisignalTypeList = self.outputDirMultisignalTypeList()
        for direction,multisignalType in outputDirMultisignalTypeList:
            outputDirMultisignal = multisignalType.createNoNameCheck(**outputMultisignal.nameValDict)
            # ensure that if there is only one output value, a transition isn't necessary
            noChoiceMs = multisignalType.noChoiceMultisignal()
            outputDirMultisignal = outputDirMultisignal.nameUnionDuplicateSignalAllowed(noChoiceMs)
            completeOutputMultisignal = completeOutputMultisignal.nameUnionDuplicateSignalAllowed(outputDirMultisignal)
            glueLabel = str(outputDirMultisignal)
            outputNeighborhood = self.outputNeighborhood(direction)
            glueAnnotation = outputNeighborhood.glueAnnotation
            glue = (glueLabel + glueAnnotation, outputNeighborhood.strength)
            if   direction == N: northglue = glue
            elif direction == S: southglue = glue
            elif direction == W: westglue = glue
            elif direction == E: eastglue = glue
            else: raise ValueError('{0} is not a valid direction'.format(direction))
        
        # set property values
        tilePropertiesToSet = {}
        for propertyName,propertyFunction in self.propertyFunctions.items():
            propertyValue = propertyFunction.apply(inputMultisignal, completeOutputMultisignal)
            tilePropertiesToSet[propertyName] = propertyValue
        
        # place calculated glues in tileProperties to set along with other properties
        gluePropertiesToSet = {}
        for glue,gluename in zip((northglue,southglue,eastglue,westglue), 
                                 ('northglue','southglue','eastglue','westglue')):
            if glue:
#                if gluename in self.tileTypeArgs:
#                    raise ValueError('glue {0} is already hardcoded into {1} and cannot be calculated'.format(gluename, self)) 
                gluePropertiesToSet[gluename] = glue
        
        tilePropertiesToSet.update(self.tileTypeArgs)
        tilePropertiesToSet.update(gluePropertiesToSet)
        tile = Tile(name = tilename, **(tilePropertiesToSet)
        )
        return tile
    
    def getStrengthErrors(self):
        """Return list of errors associated with strength mismatches.
        
        If the number of input sides is 1, the strength must be 2.
        If the number of input sides is 2, the strength of each must be 1.
        If the number of input sides is > 2, this alone constitutes an error. """
        errors = []
        if self.numInputSides() == 1:
            joins = self.inputSides()[0]
            for join in joins:
                if join.strength != 2:
                    errors.append(StrengthError(self, join))
        if self.numInputSides() == 2:
            for joins in self.inputSides.values():
                for join in joins:
                    if join.strength != 1:
                        errors.append(StrengthError(self, join))
        if self.numInputSides() > 2:
            errors.append(TooManyInputSidesError(self))
        return errors
    
    def isValidInputMultisignal(self, ms):
        """Indicates whether ms is a valid multisignal for this TileTemplate's total input multisignal type."""
        inputMultisignalType = self.inputMultisignalType()
        return inputMultisignalType.isValidMultisignal(ms)
    
    def inputMultisignalType(self, direction=None):
        """If direction is None or unspecified, returns multisignal type of all inputs.
        if direction is one of Direction enum values, then returns the MultisignalType
        of just that direction (including nondeterministic inputs)"""
        mst = MultisignalType(VERBOSE_SIGNALS)
        mstDict = self.inputDirMultisignalTypeDict()
        if direction is None:
            for d in mstDict.keys():
                mst = mst.nameUnion(mstDict[d])
        else:
            mst = mstDict[direction]
        return mst
        
    def addTransition(self, inputs=(), outputs=(), function=None, 
                      expression=None, table=None):
        """Add a transition computing output signal(s) to this TileTemplate.
        
        inputs and outputs are lists of strings specifying the names of the input
        and output signal names, in the order in which the function transition 
        receives them as arguments or returns them in a tuple."""
        inputs = ensureIsSequence(inputs)
        outputs = ensureIsSequence(outputs)
        self.transitions.append(Transition(inputs=inputs, outputs=outputs, 
                    function=function, expression=expression, table=table))
        
    def removeTransition(self, outputs):
        """Removes the transition with the given list of output signal names.
        
        If no such tuple of output signal names exists, raises a TODO"""
        if not isSequence(outputs):
            outputs = (outputs,)
        else:
            outputs = tuple(outputs)
        newTransitions = filter(lambda t: t.outputNames != outputs, self.transitions)
        if len(newTransitions) != len(self.transitions) - 1:
            raise NonexistentTransitionError(outputs, self.transitions)
        else:
            self.transitions = newTransitions
    
    def setPropertyFunction(self, name, inputs=[], outputs=[], function=None, expression=None):
        self.propertyFunctions[name] = PropertyFunction(inputs=inputs, 
            outputs=outputs, function=function, expression=expression)
        
    def setLabelFunction(self, inputs=[], outputs=[], function=None, expression=None):
        self.setPropertyFunction('label', inputs, outputs, function, expression)
        
    def setConcentrationFunction(self, inputs=[], outputs=[], function=None, expression=None):
        self.setPropertyFunction('concentration', inputs, outputs, function, expression)
        
    def setTilecolorFunction(self, inputs=[], outputs=[], function=None, expression=None):
        self.setPropertyFunction('tilecolor', inputs, outputs, function, expression)
        
    def setTextcolorFunction(self, inputs=[], outputs=[], function=None, expression=None):
        self.setPropertyFunction('textcolor', inputs, outputs, function, expression)
        
    def removePropertyFunction(self, name):
        del self.propertyFunctions[name]
    
    def numInputSides(self):
        if not self.inputJoins:
            return 0
        sides = set(join.neighborhood.direction for join in self.inputJoins)
        return len(sides)
    
    def numOutputSides(self):
        if not self.outputJoins:
            return 0
        sides = set(join.neighborhood.direction for join in self.outputJoins)
        return len(sides)
    

def ensureIsSequence(obj):
    if isSequence(obj):
        return tuple(obj)
    else:
        return (obj,)

class Neighborhood:
    def __init__(self, strength, direction):
        self.strength = strength
        self.direction = direction
        self.glueAnnotation = None
        self.multisignalType = None
        self.joins = []

    def addJoin(self, join):
        self.joins.append(join)
        self._update()
    
    def addJoins(self, joins):
        self.joins.extend(joins)
        self._update()

    def _update(self):
        """Updates glueAnnotation and multisignalType to reflect all joins."""
        self.multisignalType = None
        inputTileTemplateNames = set()
        outputTileTemplateNames = set()
        for join in self.joins:
            inputTileTemplateNames.add(join.fromTileTemplate.name)
            outputTileTemplateNames.add(join.toTileTemplate.name)
            if self.multisignalType is None:
                self.multisignalType = join.multisignalType
            else:
                self.multisignalType.valueUnion(join.multisignalType)
        inputTileTemplateNames = sorted(inputTileTemplateNames)
        outputTileTemplateNames = sorted(outputTileTemplateNames)
        self.glueAnnotation = ';{0}-{1}>{2}'.format(','.join(inputTileTemplateNames),
                                                    directionShortName(self.direction),
                                                    ','.join(outputTileTemplateNames))

    def combine(self, other):
        assert self.direction == other.direction and self.strength == other.strength
        ret = Neighborhood(self.strength, self.direction)
        ret.addJoins(self.joins)
        ret.addJoins(other.joins)
        for join in ret.joins:
            join.neighborhood = ret
        return ret
        


class ChooserSet:
    # A ChooserSet contains exactly the set of TileTemplates which are in the same
    # neighborhoods for all input sides
    # set of tile templates; disjoint union of all chooser sets should be set of all tile templates in tile set template
    # needs chooser function in addition to tile templates if there are more than 1
    # chooser only needs to work for overlapping input signal values; library should make correct choice otherwise
    
    # inputMultisignalType() (nameUnion over input sides of the valueUnion over tile templates)
    
    def __init__(self, tileSetTemplate, tileTemplate):
        self.tileSetTemplate = tileSetTemplate
        self.inputMst = tileTemplate.inputMultisignalType()
        self.tileSet = list([tileTemplate])
        self.neighborhoods = tileTemplate.inputNeighborhoods()
        self.inputs = None
        self.function = None

    def inputMultisignalType(self):
        return self.inputMst

    def chooseTileTemplates(self, inputMultisignal):
        matchSet = set()
        for tt in self.tileSet:
            if tt.isValidInputMultisignal(inputMultisignal):
                matchSet.add(tt)
        if len(matchSet) == 1:
            return list(matchSet)
        elif len(matchSet) == 0:
            return []
        else:
            tileTemplatesNamesWithChoosers = set(tileTemplate.name for tileTemplate in matchSet) & set(self.tileSetTemplate.chooserFunctionDict.keys())
            if len(tileTemplatesNamesWithChoosers) == 0:
                raise(MissingChooserError(matchSet))
            elif len(tileTemplatesNamesWithChoosers) > 1:
                raise(ConflictingChooserError(matchSet))

            chooserFunction = self.tileSetTemplate.chooserFunctionDict[tileTemplatesNamesWithChoosers.pop()]
            retList = chooserFunction.apply(inputMultisignal)

            #  Make sure we're returning tileTemplate objects instead of strings
            def ensureNameIsTileTemplate(tt):
                if isinstance(tt, str):
                    return self.findTileTemplateByName(tt)
                elif tt not in matchSet:
                    raise InvalidChooserTileTemplateError(matchSet, tileTemplate=tt)
                else:
                    return tt

            retList = (ensureNameIsTileTemplate(tt) for tt in retList)

            return retList

    def findTileTemplateByName(self, name):
        for tt in self.tileSet:
            if name == tt.name:
                return tt
        raise InvalidChooserTileTemplateError(self.tileSet, name=name)

    def belongsInSet(self, tileTemplate):
        return self.neighborhoods == tileTemplate.inputNeighborhoods()
    
    def isInSet(self, tileTemplate):
        return (tileTemplate in self.tileSet)

    def addTileTemplate(self, tileTemplate):
        if not self.belongsInSet(tileTemplate):
            raise ValueError('{0} tile template does not belong in chooser set {1}'.format(tileTemplate, self))
        # This will throw an exception if the signal names are not identical
#TODO - make sure not to include auxiliary inputs
        self.addMultisignalValues(tileTemplate)
        self.tileSet.append(tileTemplate)
    
    def addMultisignalValues(self, tileTemplate):
        self.inputMst = self.inputMst.valueUnion(tileTemplate.inputMultisignalType())

    def removeTileTemplate(self, tileTemplate):      
        self.inputMst = None
        self.tileSet.remove(tileTemplate)

        for tt in self.tileSet:
            if self.inputMst is None:
                self.inputMst = tt.inputMultisignalType()
            else:
                self.inputMst.valueUnion(tt.inputMultisignalType())


class Join:
    # needs direction, in and out tiles/tile templates, multisignal type, neighborhood, strength

#DSD: no strength or direction here anymore because all joins in a neighborhood
# share that, so the neighborhood has that info
    def __init__(self, fromTileTemplate, toTileTemplate, neighborhood, multisignalType):
        self.fromTileTemplate = fromTileTemplate
        self.toTileTemplate = toTileTemplate
        self.neighborhood = neighborhood
        self.multisignalType = multisignalType
#MJP: TODO - should this throw an exception if the join is already in these lists?  probably.
        if self not in toTileTemplate.inputJoins:
            toTileTemplate.inputJoins.append(self)
        if self not in fromTileTemplate.outputJoins:
            fromTileTemplate.outputJoins.append(self)


class TileSetTemplate:
    # needs list of chooser sets, list of joins, and list of neighborhoods, and list of hard-coded tiles
    
    # join, unjoin, rejoin, add/remove chooser, add/remove tile template/tile, createTiles
    
    def __init__(self):
        self.hardcodedTiles = []
        self.neighborhoodSet = []
        self.chooserSets = []
        self.tilesWithoutInputs = []
        self.chooserFunctionDict = {}
        
    
    def errors(self):
        errors = []
        
#        for tt in self.tileSetTemplates():
#            errors += tt.getStrengthErrors()
#        
#        try:
#            self.createTilesAndBuildErrors()
#        except ErrorListError as errorListError:
#            errors += errorList.errors()
            
        return errors
    
#    def createTilesAndBuildErrors(self):
#        errors = self.errors()
#        if len(errors) > 0:
#            raise errors[0]
#        
#        for chooserSet in self.chooserSets:
#            errors += chooserSet.getMissingChooserErrors()
#            for inputMultisignal in chooserSet.inputMultisignalType():
#                pass
#        # for each chooser set cs
#        #   for each input multisignal ims shared by all tile templates in cs
#        #     call chooser for cs to find list [tt1,tt2,...,ttK] of tile templates to generate a tile from
#        #     for each tile template tt in [tt1,tt2,...,ttK]
#        #       generate tile from tt on input ims
    
    def addTile(self, tile):
        """Add tile to this TileSetTemplate so it will be enumerated with the
        generated tiles and the tiles that were joined to a TileTemplate.
        
        This only needs to be called if the tile is never joined to a 
        TileTemplate, simply as a way to alert the TileSetTemplate of its
        existence."""
        self.hardcodedTiles.append(tile)
    
    def ensureIsTemplate(self, tile):
        """Wrap tile in a TileTemplate if tile is not a TileTemplate already.
        
        This TileTemplate has no transitions or property functions, so it will do
        nothing interesting when TileTemplate.createTilesFromInputMultisignal is
        called, but will create a tile with the same properties as this one."""
        if isinstance(tile, TileTemplate):
            return tile
        else:
            props = dict(tile.__dict__)
            name = props['name']
            del props['name']
            tt = TileTemplate(name, **props)
#TODO - give this a single signal            tt.addAuxiliaryInput()
            return tt

    def setChooser(self, tileTemplate, inputs=(), function=None, expression=None, table=None):
        self.chooserFunctionDict[tileTemplate.name] = ChooserFunction(inputs, function, expression, table)

    def removeChooser(self, tileTemplate):
        if tileTemplate.name in self.chooserFunctionDict:
            del self.chooserFunctionDict[tileTemplate.name]
        else:
            raise NonexistentChooserSetError(tileTemplate)

    def createTiles(self):
        """Return list of all tiles this TileSetTemplate generates.
        
        If unresolved errors exist, the first such error encountered is raised."""
        errors = self.errors()
        if errors:
            raise errors[0]
        tiles = list(self.hardcodedTiles)
        for chooserSet in self.chooserSets:
            for inputMultisignal in chooserSet.inputMultisignalType():
                tileTemplates = chooserSet.chooseTileTemplates(inputMultisignal)
                for tileTemplate in tileTemplates:
                    tilesFromTileTemplate = tileTemplate.createTilesFromInputMultisignal(inputMultisignal)
                    tiles.extend(tilesFromTileTemplate)
            
        # Handle tiles with no inputs, which are typically Tiles wrapped in TileTemplates        
        emptyMs = Multisignal(())
        for tileWithoutInput in self.tilesWithoutInputs:
            tilesFromTileTemplate = tileWithoutInput.createTilesFromInputMultisignal(emptyMs)
            tiles.extend(tilesFromTileTemplate)

        return [tile for tile in tiles if tile is not None]
    

    def removeJoin(self, strength, direction, fromTileTemplate, toTileTemplate):
        #  Find the join which is to be removed
        joinToRemove = None
        for neighborhood in self.neighborhoodSet:
            if neighborhood.strength == strength and neighborhood.direction == direction:
                for j in neighborhood.joins:
                    if j.fromTileTemplate == fromTileTemplate and j.toTileTemplate == toTileTemplate:
                        joinToRemove = j
                        break
            if joinToRemove is not None:
                break
        
        if joinToRemove is not None:
            #  Remove the join from the list of joins associated with the to and from tile templates
            if joinToRemove in joinToRemove.toTileTemplate.inputJoins:
                joinToRemove.toTileTemplate.inputJoins.remove(joinToRemove)
            if joinToRemove in joinToRemove.fromTileTemplate.outputJoins:
                joinToRemove.fromTileTemplate.outputJoins.remove(joinToRemove)
            
            #  Get rid of the neighborhood that this join was in and build a new one(s)
            nbrhdToRemove = joinToRemove.neighborhood
            if nbrhdToRemove in self.neighborhoodSet:
                self.neighborhoodSet.remove(nbrhdToRemove)

            #  Remove join from the join list, then find new neighborhoods for the other joins
            nbrhdToRemove.joins.remove(joinToRemove)
            for j in nbrhdToRemove.joins:
                newNbrhd = self._getNeighborhoodForJoin(j.neighborhood.strength, j.neighborhood.direction, j.fromTileTemplate, j.toTileTemplate)
                newNbrhd.addJoin(j)

            #  Remove the 'toTileTemplate' from its chooser set and then find/create the one it should be in
            for chooserSet in self.chooserSets:
                if chooserSet.isInSet(joinToRemove.toTileTemplate):
                    chooserSet.removeTileTemplate(joinToRemove.toTileTemplate)
                    if len(chooserSet.tileSet) == 0:
                        self.chooserSets.remove(chooserSet)
                    break

            bInSet = False
            for chooserSet in self.chooserSets:
                if chooserSet.belongsInSet(joinToRemove.toTileTemplate):
                    chooserSet.addTileTemplate(joinToRemove.toTileTemplate)
                    bInSet = True
                    break
            if not bInSet:
                cs = ChooserSet(self, joinToRemove.toTileTemplate)
                self.chooserSets.append(cs)
                
            # In case the tile which was receiving the join no longer has an input, make sure it isn't lost
            if joinToRemove.toTileTemplate.numInputSides() == 0:
                self.tilesWithoutInputs.append(joinToRemove.toTileTemplate)


    def _getNeighborhoodForJoin(self, strength, direction, fromTileTemplate, toTileTemplate):
        """Find/create neigbhorhood
           Because we could be joining Tiles or TileTemplates, 
             loop over all neighborhoods to look for a fit"""
        inNbrhd = None
        outNbrhd = None
        for nbrhd in self.neighborhoodSet:
            if nbrhd.direction == direction:
                for join in nbrhd.joins:
                    if fromTileTemplate == join.fromTileTemplate:
                        inNbrhd = nbrhd
                        if strength != nbrhd.strength:
                            raise StrengthError('strengths {0} and {1} do not match'.format(
                                                strength,nbrhd.strength))
                    if toTileTemplate == join.toTileTemplate:
                        outNbrhd = nbrhd
                        if strength != nbrhd.strength:
                            raise StrengthError('strengths {0} and {1} do not match'.format(
                                                strength,nbrhd.strength))

        myNbrhd = None
        if (inNbrhd is not None) and (outNbrhd is not None) and (inNbrhd is not outNbrhd):
            # There are two disjoint neighborhoods which need to be combined to
            # form a single neighborhood as long as the multiSignalType names match
            myNbrhd = inNbrhd.combine(outNbrhd)
            self.neighborhoodSet.remove(inNbrhd)
            self.neighborhoodSet.remove(outNbrhd)
        elif inNbrhd is not None:
            # Either there is only an inNbrhd to join to, or inNbrhd and outNbrhd
            # are the same.  Either way, enter inNbrhd as long as the multiSignalType names match            
            myNbrhd = inNbrhd
        elif outNbrhd is not None:
            # There is only an outNbrhd, so enter it as long as the multiSignalType names match
            myNbrhd = outNbrhd
        else:
            # There is no applicable, existing neighborhood, so create a new one
            myNbrhd = Neighborhood(strength, direction)
            self.neighborhoodSet.append(myNbrhd)

        return myNbrhd
        

    def join(self, strength, direction, fromTileTemplate, toTileTemplate, 
             multisignalType=None, **kwargs):
        # combine multisignalType with signals specified in kwargs
        if multisignalType is None:
            multisignalType = MultisignalType(**kwargs)
        else:
            multisignalType = copy.copy(multisignalType)
            multisignalType.addSignalTypes(**kwargs)
            
        # wrap Tile in TileTemplate if necessary
        fromTileTemplate = self.ensureIsTemplate(fromTileTemplate)
        toTileTemplate = self.ensureIsTemplate(toTileTemplate)
        
        #TODO: check if either TileTemplate is new; if so, ensure name not already in use
        
        # Make sure there isn't already a join in this direction between these two tiles
        for neighborhood in self.neighborhoodSet:
            for join in neighborhood.joins:
                if (join.neighborhood.direction == direction and 
                    join.fromTileTemplate == fromTileTemplate and 
                    join.toTileTemplate == toTileTemplate):
                    raise ValueError('{0} and {1} are already joined in direction {2}'.format(
                        fromTileTemplate.name,toTileTemplate.name,direction))

        # Ensure direction of join doesn't conflict with fromTileType's current state
        if direction in fromTileTemplate.inputDirMultisignalTypeDict().keys():
            raise InputOutputSideConflictError(fromTileTemplate, direction)
        # Ensure direction of join doesn't conflict with toTileType's current state
        if oppositeDirection(direction) in toTileTemplate.outputDirMultisignalTypeDict().keys():
            raise InputOutputSideConflictError(toTileTemplate, direction)

        #  Find/create the neighborhood for this join
        myNbrhd = self._getNeighborhoodForJoin(strength, direction, fromTileTemplate, toTileTemplate)

        newJoin = Join(fromTileTemplate, toTileTemplate, myNbrhd, multisignalType)
        myNbrhd.addJoin(newJoin)

        toTileTemplates = set([join.toTileTemplate for join in myNbrhd.joins])
        for toTileTemplate in toTileTemplates:
            bInSet = False
            for chooserSet in self.chooserSets:
                if chooserSet.belongsInSet(toTileTemplate):
                    if not chooserSet.isInSet(toTileTemplate):
                        chooserSet.addTileTemplate(toTileTemplate)
                    else:
                        chooserSet.addMultisignalValues(toTileTemplate)
                    bInSet = True
                    break
            if not bInSet:
                cs = ChooserSet(self, toTileTemplate)
                self.chooserSets.append(cs)

            for chooserSet in self.chooserSets[:]:
                if chooserSet.isInSet(toTileTemplate):
                    if not chooserSet.belongsInSet(toTileTemplate):
                        chooserSet.removeTileTemplate(toTileTemplate)
                        if len(chooserSet.tileSet) == 0:
                            self.chooserSets.remove(chooserSet)

        # In case the tile which is outputting to the join doesn't have an input, make sure it isn't lost
        if fromTileTemplate.numInputSides() == 0:
            if fromTileTemplate not in self.tilesWithoutInputs:
                self.tilesWithoutInputs.append(fromTileTemplate)
        
        # If toTileTemplate was in the list of tile templates without inputs, remove it
        if toTileTemplate in self.tilesWithoutInputs:
            self.tilesWithoutInputs.remove(toTileTemplate)
        
        return newJoin

 
